from abc import ABC, abstractmethod

from tktl.core.clients.taktile import TaktileClient
from tktl.core.config import settings
from tktl.core.loggers import SdkLogger
from tktl.core.t import ServiceT
from tktl.login import login


class Client(ABC):
    TRANSPORT: str = ServiceT

    def __init__(self, api_key: str, repository_name: str, logger=SdkLogger()):
        login(api_key)
        self.taktile_client = TaktileClient(api_url=settings.DEPLOYMENT_API_URL)
        self.repository_name = repository_name
        self.logger = logger

    @abstractmethod
    def predict(self, *args, **kwargs):
        ...

    @abstractmethod
    def list_endpoints(self):
        ...

    @abstractmethod
    def list_deployments(self):
        ...

    @abstractmethod
    def get_sample_data(self):
        ...

    @abstractmethod
    def get_schema(self, *args, **kwargs):
        ...



from tktl.core.serializers.arrow.factory import deserialize_arrow, serialize_arrow
from tktl.core.serializers.rest.factory import deserialize_rest, serialize_rest
from tktl.core.serializers.rest.typed import to_pydantic



import warnings
from typing import Iterator, Sequence, Union

import pandas
import pandas.api.types as ptypes

from tktl.core.exceptions import exceptions
from tktl.core.loggers import CliLogger

logger = CliLogger()


def validate_outputs(preds) -> Union[bool, Iterator]:
    if not isinstance(preds, pandas.Series):
        preds = pandas.Series(preds)
    if not ptypes.is_numeric_dtype(pandas.Series(preds)):
        logger.warning("Function return types are not numeric, profiling won't happen")
        return False
    return True


def validate_binary(preds):
    if not 0 <= min(preds):
        raise exceptions.ValidationException(
            "Function output cannot be negative for endpoint kind binary"
        )

    if not max(preds) <= 1:
        raise exceptions.ValidationException(
            "Function output cannot exceed 1 for endpoint kind binary"
        )
    return True


def validate_shapes(
    x_frame: pandas.DataFrame, y_series: pandas.Series, preds: Sequence
):
    return len(x_frame) == len(preds) == len(y_series)


def data_frame_convertible(inputs):
    if len(inputs) > 1e6:
        warnings.warn(
            f"inputs is very large (N={len(inputs)}). Please consider using a smaller reference dataset."
        )
    if isinstance(inputs, pandas.DataFrame):
        return True
    try:
        pandas.DataFrame(inputs)
    except Exception as e:
        logger.warning(
            f"Could not convert inputs to pd.DataFrame: {repr(e)}, profiling won't happen"
        )
        return False
    return True


def series_convertible(y, type_cast=None):
    if isinstance(y, pandas.Series):
        return True
    try:
        y = pandas.Series(y)
        if type_cast:
            y = y.astype(type_cast)
        y.name = y.name or "Outcome"

    except Exception as e:
        logger.warning(
            f"Could not convert y to pandas series: {repr(e)}, profiling won't happen"
        )
        return False
    return True



from typing import Any, Callable, List

from pydantic import UUID4

from .. import loggers as sdk_logger
from ..schemas.repository import Endpoint, Repository, RepositoryDeployment
from ..t import Resources

_logger = sdk_logger.SdkLogger()

default_comp = lambda x, y: x == y


def filter_prop(
    resources: List[Resources], prop_name: str, value: Any, key: Callable = default_comp
):
    if value is not None:
        return [d for d in resources if key(value, getattr(d, prop_name))]
    return resources


def filter_empty_and_duplicate_resources(resources: List[Resources]):
    non_empty = [r for r in resources if r.id]
    # can't use set() since BaseModels are not hashable
    return [x for i, x in enumerate(non_empty) if i == non_empty.index(x)]


def filter_deployments(
    deployments: List[RepositoryDeployment],
    deployment_id: UUID4,
    commit_sha: str = None,
    branch_name: str = None,
    status_name: str = None,
) -> List[RepositoryDeployment]:
    if deployment_id:
        deployments = [d for d in deployments if d.id == deployment_id]

        if deployments:
            return deployments
        else:
            _logger.warning("No deployments with matching id found")

    for n, v in zip(
        ["commit_hash", "branch_name", "status"], [commit_sha, branch_name, status_name]
    ):
        if n == "commit_hash" and commit_sha:
            if len(commit_sha) < 40:
                comp = lambda x, y: y.startswith(x)
            else:
                comp = default_comp
        else:
            comp = default_comp
        deployments = filter_prop(deployments, prop_name=n, value=v, key=comp)

    return filter_empty_and_duplicate_resources(deployments)


def filter_repositories(
    repositories: List[Repository],
    repository_id: UUID4,
    repository_name: str = None,
    repository_owner: str = None,
) -> List[Repository]:
    if repository_id:
        repositories = [r for r in repositories if r.id == repository_id]
        if repositories:
            return repositories
        else:
            _logger.warning("No repositories with matching id found")

    for n, v in zip(
        ["repository_name", "repository_owner"], [repository_name, repository_owner]
    ):
        repositories = filter_prop(repositories, prop_name=n, value=v)
    return filter_empty_and_duplicate_resources(repositories)


def filter_endpoints(
    endpoints: List[Endpoint], endpoint_id: UUID4, endpoint_name: str = None
) -> List[Endpoint]:
    if endpoint_id:
        endpoints = [r for r in endpoints if r.id == endpoint_id]
        if endpoints:
            return endpoints
        else:
            _logger.warning("No endpoints with matching id found")

    for n, v in zip(["endpoint_name"], [endpoint_name]):
        endpoints = filter_prop(endpoints, prop_name=n, value=v)
    return filter_empty_and_duplicate_resources(endpoints)



from datetime import datetime
from typing import Dict, List, Optional, Union

from beautifultable import BeautifulTable
from pydantic import UUID4, BaseModel

from tktl.core import ExtendedEnum
from tktl.core.utils import flatten


class AccessKind(str, ExtendedEnum):

    # see also corresponding AccessKind on t-api
    OWNER = "owner"
    VIEWER = "viewer"


class TablePrintableBaseModelMixin:
    def table_repr(self, subset: List[str] = None) -> Dict:
        ...


class Endpoint(BaseModel):
    id: Optional[UUID4] = None
    name: Optional[str] = None
    deployment_id: Optional[UUID4] = None

    def table_repr(self, subset=None):
        as_dict = self.dict()
        as_dict["Name"] = as_dict.pop("name")
        as_dict["Deployment ID"] = str(as_dict.pop("deployment_id"))
        return as_dict


class RepositoryDeployment(BaseModel, TablePrintableBaseModelMixin):
    id: UUID4
    created_at: datetime
    status: str
    public_docs_url: Optional[str]
    service_type: Optional[str]
    instance_type: Optional[str]
    replicas: Optional[int]
    branch_name: str
    major_version: int
    minor_version: int
    commit_hash: str
    endpoints: List[Endpoint]

    def table_repr(self, subset=None):
        endpoint_names = [e.name for e in self.endpoints]
        as_dict = self.dict(exclude={"service_type", "endpoints"})
        as_dict[
            "Branch @ Commit"
        ] = f"{as_dict.pop('branch_name')} @ {as_dict.pop('commit_hash')[0:7]}"
        as_dict["Status"] = as_dict.pop("status")
        as_dict["Endpoints"] = ", ".join(endpoint_names)
        as_dict["Created At"] = str(as_dict.pop("created_at"))
        as_dict[
            "Version"
        ] = f"{as_dict.pop('major_version')}.{as_dict.pop('minor_version')}"
        as_dict["Instance Type"] = as_dict.pop("instance_type")
        as_dict["Replicas"] = as_dict.pop("replicas")
        as_dict["REST Docs URL"] = _format_http_url(as_dict.pop("public_docs_url"))
        if subset:
            return {k: v for k, v in as_dict.items() if k in subset}
        return as_dict


class Repository(BaseModel, TablePrintableBaseModelMixin):
    id: UUID4
    ref_id: int
    repository_name: str
    repository_owner: str
    repository_description: Optional[str] = None
    access: AccessKind
    deployments: List[RepositoryDeployment]

    def table_repr(self, subset=None):
        deployment_ids = [str(e.id) for e in self.deployments]
        as_dict = self.dict(exclude={"ref_id", "deployments"})
        as_dict["Deployments"] = "\n".join(deployment_ids)
        as_dict["Access"] = f"{as_dict.pop('access').value}"
        as_dict[
            "Name"
        ] = f"{as_dict.pop('repository_owner')}/{as_dict.pop('repository_name')}"
        desc = as_dict.pop("repository_description")
        as_dict["Description"] = f"{desc[0:20] + '...' if desc else '-'}"
        if subset:
            return {k: v for k, v in as_dict.items() if k in subset}
        return as_dict


class RepositoryList(BaseModel):
    __root__: List[Repository]

    def get_repositories(self) -> List[Repository]:
        return flatten(
            [
                sum(
                    [
                        len(e.endpoints) if len(e.endpoints) > 0 else 1
                        for e in r.deployments
                    ]
                )
                * [r]
                for r in self.__root__
            ]
        )

    def get_endpoints(self) -> List[Endpoint]:
        return flatten(
            [
                flatten(
                    [
                        [e for e in d.endpoints] if d.endpoints else [Endpoint()]
                        for d in r.deployments
                    ]
                )
                for r in self.__root__
            ]
        )

    def get_deployments(self) -> List[RepositoryDeployment]:
        return flatten(
            flatten(
                [
                    (len(d.endpoints) if len(d.endpoints) > 0 else 1) * [d]
                    for d in r.deployments
                ]
                for r in self.__root__
            )
        )

    def __str__(self):
        table = BeautifulTable(maxwidth=250)
        rows = self.total_rows()
        table.columns.header = [
            "Repository",
            "Branch @ Commit",
            "Deployment Status",
            "REST Docs URL",
            "Version",
            "Created At",
            "Endpoint Name",
            "Endpoint ID",
        ]
        table.rows.header = [str(i) for i in range(rows)]
        table.columns[0] = [
            f"{r.repository_owner}/{r.repository_name}" for r in self.get_repositories()
        ]
        table.columns[1] = [
            f"{d.branch_name} @ {d.commit_hash}" for d in self.get_deployments()
        ]
        table.columns[2] = [f"{d.status}" for d in self.get_deployments()]
        table.columns[3] = [
            f"{_format_http_url(d.public_docs_url)}" for d in self.get_deployments()
        ]
        table.columns[4] = [
            f"{d.major_version}.{d.minor_version}" for d in self.get_deployments()
        ]
        table.columns[5] = [f"{d.created_at}" for d in self.get_deployments()]
        table.columns[6] = [e.name if e.name else "" for e in self.get_endpoints()]
        table.columns[7] = [e.id if e.id else "" for e in self.get_endpoints()]
        return table.__str__()

    def total_rows(self):
        return sum(
            [len([len(d.endpoints) for d in r.deployments]) for r in self.__root__]
        )


class ReportResponse(BaseModel):
    deployment_id: UUID4
    endpoint_name: str
    report_type: str
    chart_name: Optional[str] = None
    variable_name: Optional[str] = None
    value: Union[List, Dict]


def _format_http_url(url):
    return f"https://{url}" if (url and url != "UNAVAILABLE") else "UNAVAILABLE"


def _format_grpc_url(url):
    return f"grpc+tcp://{url}:443" if (url and url != "UNAVAILABLE") else "UNAVAILABLE"



import functools
import itertools
import os
import time

from tktl.core.exceptions import WrongPathError


def concatenate_urls(fst_part, snd_part):
    fst_part = fst_part if not fst_part.endswith("/") else fst_part[:-1]
    template = "{}{}" if snd_part.startswith("/") else "{}/{}"
    concatenated = template.format(fst_part, snd_part)
    return concatenated


class PathParser(object):
    LOCAL_DIR = 0
    LOCAL_FILE = 1
    GIT_URL = 2
    S3_URL = 3

    @classmethod
    def parse_path(cls, path):
        if cls.is_local_dir(path):
            return cls.LOCAL_DIR

        if cls.is_local_zip_file(path):
            return cls.LOCAL_FILE

        if cls.is_git_url(path):
            return cls.GIT_URL

        if cls.is_s3_url(path):
            return cls.S3_URL

        raise WrongPathError("Given path is neither local path, nor valid URL")

    @staticmethod
    def is_local_dir(path):
        return os.path.exists(path) and os.path.isdir(path)

    @staticmethod
    def is_local_zip_file(path):
        return os.path.exists(path) and os.path.isfile(path) and path.endswith(".zip")

    @staticmethod
    def is_git_url(path):
        return (
            not os.path.exists(path)
            and path.endswith(".git")
            or path.lower().startswith("git:")
        )

    @staticmethod
    def is_s3_url(path):
        return not os.path.exists(path) and path.lower().startswith("s3:")


def lru_cache(timeout: int, maxsize: int = 128, typed: bool = False):
    def wrapper_cache(func):
        func = functools.lru_cache(maxsize=maxsize, typed=typed)(func)
        func.delta = timeout * 10 ** 9
        func.expiration = time.monotonic_ns() + func.delta

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            if time.monotonic_ns() >= func.expiration:
                func.cache_clear()
                func.expiration = time.monotonic_ns() + func.delta
            return func(*args, **kwargs)

        wrapped_func.cache_info = func.cache_info
        wrapped_func.cache_clear = func.cache_clear
        return wrapped_func

    return wrapper_cache


flatten = lambda x: list(itertools.chain.from_iterable(x))



from pyarrow.flight import (
    ClientAuthHandler,
    FlightClient,
    FlightDescriptor,
    FlightInfo,
    Ticket,
)

from tktl.core.clients import Client
from tktl.core.exceptions import AuthenticationError
from tktl.core.schemas.repository import _format_grpc_url
from tktl.core.serializers import deserialize_arrow, serialize_arrow
from tktl.core.t import ServiceT


class ApiKeyClientAuthHandler(ClientAuthHandler):
    """An example implementation of authentication via ApiKey."""

    def __init__(self, api_key: str):
        super(ApiKeyClientAuthHandler, self).__init__()
        self.api_key = api_key

    def authenticate(self, outgoing, incoming):
        outgoing.write(self.api_key)
        self.api_key = incoming.read()

    def get_token(self):
        return self.api_key


class ArrowFlightClient(Client):

    TRANSPORT = ServiceT.GRPC

    def __init__(self, api_key: str, repository_name: str, local: bool = False):
        super().__init__(api_key=api_key, repository_name=repository_name)
        self._location = None
        self._endpoint = None
        self._client = None
        self.local = local

    def list_endpoints(self):
        return {
            e["endpoint"].name: _format_grpc_url(e["deployment"].public_docs_url)
            for e in self.taktile_client.get_endpoints_for_repository(
                repository=self.repository_name
            )
        }

    @property
    def local_endpoint(self):
        return "grpc+tcp://0.0.0.0:5005"

    def list_deployments(self):
        pass

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value):
        self._location = value

    @property
    def endpoint(self):
        return self._endpoint

    @endpoint.setter
    def endpoint(self, value):
        self._endpoint = value

    @property
    def client(self):
        return self._client

    @client.setter
    def client(self, client: FlightClient):
        self._client = client

    def list_commands(self):
        return self.client.list_actions()

    def __set_endpoint_and_location(self, endpoint_name: str):
        self.endpoint = endpoint_name
        if self.local:
            self.location = self.local_endpoint
        else:
            self.location = self.list_endpoints()[endpoint_name]

    def authenticate(self, endpoint_name: str):
        self.__set_endpoint_and_location(endpoint_name)
        client = FlightClient(location=self.location)
        client.authenticate(
            ApiKeyClientAuthHandler(api_key=self.taktile_client.api_key)
        )
        self.client = client

    def predict(self, inputs):
        table = serialize_arrow(inputs)
        descriptor = self.get_flight_info(command_name=str.encode(self.endpoint))
        writer, reader = self.client.do_exchange(descriptor.descriptor)
        with writer:
            writer.begin(table.schema)
            writer.write_table(table)
            writer.done_writing()
            table = reader.read_all()
            return deserialize_arrow(table)

    def get_sample_data(self):
        if not self.endpoint:
            raise AuthenticationError(
                "Please authenticate against a specific endpoint first"
            )
        x_info = self.client.do_get(Ticket(ticket=str.encode(f"{self.endpoint}__X")))
        y_info = self.client.do_get(Ticket(ticket=str.encode(f"{self.endpoint}__y")))
        return (
            deserialize_arrow(x_info.read_all()),
            deserialize_arrow(y_info.read_all()),
        )

    def get_schema(self):
        if not self.endpoint:
            raise AuthenticationError(
                "Please authenticate against a specific endpoint first"
            )
        info = self.get_flight_info(str.encode(self.endpoint))
        return info.schema

    def get_flight_info(self, command_name: bytes) -> FlightInfo:
        descriptor = FlightDescriptor.for_command(command_name)
        return self.client.get_flight_info(descriptor)



from pydantic import ValidationError

from tktl.core.exceptions.exceptions import (
    MissingDocker,
    NoContentsFoundException,
    UserRepoValidationException,
)
from tktl.core.loggers import CliLogger
from tktl.core.managers.docker import DockerManager
from tktl.core.managers.project import ProjectManager
from tktl.core.schemas.project import ProjectValidationOutput
from tktl.core.validation.outputs import (
    ConfigFileValidationFailure,
    ProjectValidationFailure,
)

logger = CliLogger()


def validate_project_config(path: str):
    try:
        ProjectManager.validate_project_config(path)
    except ValidationError as e:
        validation_output = ProjectValidationOutput(
            title=ConfigFileValidationFailure.title,
            summary=ConfigFileValidationFailure.summary,
            text=ConfigFileValidationFailure.format_step_results(validation_errors=e),
        )
        log_failure(validation_output)
        return
    except (NoContentsFoundException, UserRepoValidationException) as e:
        validation_output = ProjectValidationOutput(
            title=ProjectValidationFailure.title,
            summary=ProjectValidationFailure.summary,
            text=ProjectValidationFailure.format_step_results(validation_errors=e),
        )
        log_failure(validation_output)
        return
    logger.log(f"Project scaffolding is valid!", color="green")


def log_failure(validation_output: ProjectValidationOutput):
    logger.log(
        f"Project scaffolding is invalid: {validation_output.title}", color="red"
    )
    logger.log(validation_output.summary, color="red", err=True)
    logger.log(validation_output.text, color="red", err=True)


def build_image(path: str):
    dm = DockerManager(path)

    logger.log("Patching docker file...")
    dm.patch_docker_file()
    logger.log("Building docker image...")
    image = dm.build_image()
    dm.remove_patched_docker_file()

    return dm, image


def status_output(status, logs):
    if status["StatusCode"] == 0:
        logger.log("Success", color="green")
    else:
        logger.log("Error", color="red")
        print(logs.decode())


def validate_import(path: str):
    try:
        dm, image = build_image(path)
        logger.log("Testing import of tktl...")
        status, logs = dm.test_import(image)
        status_output(status, logs)
    except MissingDocker:
        logger.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def validate_unittest(path: str):
    try:
        dm, image = build_image(path)
        logger.log("Running unittests...")
        status, logs = dm.test_unittest(image)
        status_output(status, logs)
    except MissingDocker:
        logger.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def validate_profiling(path: str):
    try:
        dm, image = build_image(path)
        logger.log("Running profiling (this can take a while)...")
        status, logs = dm.test_profiling(image)
        status_output(status, logs)
    except MissingDocker:
        logger.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def validate_integration(path: str):
    try:
        dm, image = build_image(path)
        logger.log("Waiting for service to start...")
        resp, output = dm.test_integration(image)

        if resp is None:
            logger.log("Could not access endpoint", color="red")
            print(output.decode())
        elif resp.status_code != 200:
            logger.log(f"Response status code {resp.status_code}", color="red")
            print(output.decode())
        else:
            logger.log("Success", color="green")

    except MissingDocker:
        logger.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )


def validate_all(path: str):
    validate_project_config(path)
    validate_import(path)
    validate_unittest(path)
    validate_profiling(path)
    validate_integration(path)



from enum import Enum
from typing import Iterable, List, Set


class ExtendedEnum(Enum):
    @classmethod
    def set(cls) -> Set:
        return set(map(lambda c: c.value, cls))

    @classmethod
    def list(cls) -> List:
        return list(map(lambda c: c.value, cls))

    @classmethod
    def names(cls: Iterable) -> List:
        return list(map(lambda c: c.name, cls))



from collections import defaultdict
from typing import Dict, List, Union

from pydantic import UUID4

from .. import loggers as sdk_logger
from ..config import settings
from ..exceptions import TaktileSdkError
from ..managers.auth import AuthConfigManager
from ..schemas.deployment import DeploymentBase
from ..schemas.repository import (
    Endpoint,
    Repository,
    RepositoryDeployment,
    RepositoryList,
)
from ..utils import lru_cache
from .http_client import API, interpret_response
from .utils import filter_deployments, filter_endpoints, filter_repositories

_logger = sdk_logger.SdkLogger()


class TaktileClient(API):
    SCHEME: str

    def __init__(self, api_url, logger=sdk_logger.MuteLogger()):
        """
        Base class. All client classes inherit from it.
        """
        super().__init__(api_url, logger=logger)
        self.api_url = api_url
        self.api_key = AuthConfigManager.get_api_key()
        self.logger = logger

    @lru_cache(timeout=50, typed=False)
    def __get_repositories(self) -> RepositoryList:
        response = self.get(f"{settings.API_V1_STR}/models")
        return interpret_response(response=response, model=RepositoryList)

    def _get_repositories(self) -> RepositoryList:
        repositories = self.__get_repositories()
        if not repositories:
            raise TaktileSdkError(
                "No repositories found on your account, so no resources can be fetched"
            )
        return repositories

    def get_deployments(
        self,
        deployment_id: UUID4 = None,
        commit_sha: str = None,
        branch_name: str = None,
        status_name: str = None,
        return_all: bool = False,
    ) -> List[RepositoryDeployment]:
        deployments = self._get_repositories().get_deployments()
        if return_all:
            return deployments
        return filter_deployments(
            deployments,
            deployment_id=deployment_id,
            commit_sha=commit_sha,
            branch_name=branch_name,
            status_name=status_name,
        )

    def get_repositories(
        self,
        repository_id: UUID4 = None,
        repository_name: str = None,
        repository_owner: str = None,
        return_all: bool = False,
    ) -> List[Repository]:
        repositories = self._get_repositories().get_repositories()
        if return_all:
            return repositories
        return filter_repositories(
            repositories,
            repository_id=repository_id,
            repository_name=repository_name,
            repository_owner=repository_owner,
        )

    def get_endpoints(
        self,
        endpoint_id: UUID4 = None,
        endpoint_name: str = None,
        return_all: bool = False,
    ) -> List[Endpoint]:
        endpoints = self._get_repositories().get_endpoints()
        if return_all:
            return endpoints
        return filter_endpoints(
            endpoints, endpoint_id=endpoint_id, endpoint_name=endpoint_name
        )

    def get_endpoints_for_repository(
        self, repository: str
    ) -> List[Dict[str, Union[Endpoint, RepositoryDeployment]]]:
        repos = self._get_repositories()
        if not repos:
            raise TaktileSdkError("No repos found")
        repo_models = repos.get_repositories()
        repo_endpoints = repos.get_endpoints()
        repo_deployments = repos.get_deployments()
        mapping = defaultdict(list)
        for repo, endpoint, deployment in zip(
            repo_models, repo_endpoints, repo_deployments
        ):
            mapping[f"{repo.repository_owner}/{repo.repository_name}"].append(
                {"endpoint": endpoint, "deployment": deployment}
            )
        if repository not in mapping.keys():
            raise TaktileSdkError(
                f"Repository {repository} not found. Available repos: {list(mapping.keys())}"
            )
        return mapping[repository]

    def delete_deployment(self, deployment_id: UUID4) -> DeploymentBase:
        response = self.delete(f"{settings.API_V1_STR}/deployments/{deployment_id}")
        return interpret_response(response=response, model=DeploymentBase)



import json
import os

from openapi_schema_pydantic import OpenAPI

from tktl.core.clients import Client
from tktl.core.clients.http_client import API
from tktl.core.schemas.repository import _format_http_url
from tktl.core.t import ServiceT


class RestClient(Client):
    def get_schema(self, *args, **kwargs):
        pass

    TRANSPORT = ServiceT.REST

    def __init__(self, api_key: str, repository_name: str, local: bool = False):
        super().__init__(api_key=api_key, repository_name=repository_name)
        self._location = None
        self._endpoint = None
        self._client = None
        self.local = local

    def predict(self, inputs):
        response = self.client.post(
            url=f"model/{self.endpoint}", data=json.dumps(inputs)
        )
        return response.json()

    def list_endpoints(self):
        return {
            e["endpoint"].name: _format_http_url(e["deployment"].public_docs_url)
            for e in self.taktile_client.get_endpoints_for_repository(
                repository=self.repository_name
            )
        }

    def __set_endpoint_and_location(self, endpoint_name: str):
        self.endpoint = endpoint_name
        if self.local:
            self.location = self.local_endpoint
        else:
            self.location = self.list_endpoints()[endpoint_name]

    def authenticate(self, endpoint_name: str):
        self.__set_endpoint_and_location(endpoint_name)
        client = API(api_url=self.location)
        self.client = client

    @property
    def client(self) -> API:
        return self._client

    @client.setter
    def client(self, client: API):
        self._client = client

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value):
        self._location = value

    @property
    def endpoint(self):
        return self._endpoint

    @endpoint.setter
    def endpoint(self, value):
        self._endpoint = value

    @property
    def local_endpoint(self):
        return "http://0.0.0.0:8080"

    def list_deployments(self):
        pass

    def get_sample_data(self):
        schema = self.client.get(url="openapi.json")
        openapi = OpenAPI.parse_obj(schema.json())
        request_reference, response_reference = get_endpoint_model_reference(
            openapi, endpoint=self.endpoint
        )
        sample_input = openapi.components.schemas[request_reference].example
        if not sample_input:
            self.logger.warning("No sample input found for endpoint")
        sample_output = openapi.components.schemas[response_reference].example
        if not sample_output:
            self.logger.warning("No sample output found for endpoint")
        return sample_input, sample_output


def get_endpoint_model_reference(openapi: OpenAPI, endpoint: str):
    for k in openapi.paths.keys():
        if f"/model/{endpoint}" == k:
            request_ref = (
                openapi.paths[k]
                .post.requestBody.content["application/json"]
                .media_type_schema.ref
            )
            response_ref = (
                openapi.paths[k]
                .post.responses["200"]
                .content["application/json"]
                .media_type_schema.ref
            )
            return os.path.basename(request_ref), os.path.basename(response_ref)
    return None



from collections.abc import Sequence
from functools import singledispatch
from typing import Any, Type

import numpy
import pandas
import pyarrow

from tktl.core import serializers as ser
from tktl.core.exceptions import UnsupportedInputTypeException
from tktl.core.loggers import CliLogger
from tktl.core.serializers import arrow
from tktl.core.serializers.base import ObjectSerializer

logger = CliLogger()


def deserialize_arrow(model_input: pyarrow.Table):
    cls_name = model_input.schema.metadata[b"CLS"]
    func = getattr(getattr(arrow, cls_name.decode("utf-8")), "deserialize")
    return func(model_input)


@singledispatch
def serialize_arrow(model_input):
    return arrow.BinarySerializer.serialize(model_input)


@serialize_arrow.register(Sequence)
def _(model_input):
    return _do_serialize_with_fallback(
        arrow.SequenceSerializer, model_input=model_input
    )


@serialize_arrow.register(dict)
def _(model_input):
    return _do_serialize_with_fallback(
        arrow.SequenceSerializer, model_input=model_input
    )


@serialize_arrow.register(numpy.ndarray)
def _(model_input):
    return _do_serialize_with_fallback(arrow.ArraySerializer, model_input=model_input)


@serialize_arrow.register(pandas.DataFrame)
def _(model_input):
    return _do_serialize_with_fallback(
        arrow.DataFrameSerializer, model_input=model_input
    )


@serialize_arrow.register(pandas.Series)
def _(model_input):
    return _do_serialize_with_fallback(arrow.SeriesSerializer, model_input=model_input)


def _do_serialize_with_fallback(
    serializer_cls: Type[ObjectSerializer], model_input: Any
):
    try:
        return serializer_cls.serialize(value=model_input)
    except UnsupportedInputTypeException:
        logger.warning(
            "Invalid input supported, will use binary representation for object"
        )
    return arrow.BinarySerializer.serialize(model_input)



import os
import time

import docker
import requests
from requests.exceptions import RequestException

from tktl.core.exceptions.exceptions import MissingDocker

TESTING_DOCKERFILE = "Dockerfile.taktile-cli-testing"


class DockerManager:
    def __init__(self, path):
        try:
            self._client = docker.from_env()
            self._path = path
        except docker.errors.DockerException as err:
            raise MissingDocker from err

    def get_docker_file(self) -> str:
        with open(os.path.join(self._path, "Dockerfile")) as fp:
            return fp.read()

    def patch_docker_file(self, output: str = TESTING_DOCKERFILE):
        """patch_docker_file
        
        Remove the line that does the profiling
        """
        with open(os.path.join(self._path, "Dockerfile")) as fp:
            contents = [
                line
                for line in fp.readlines()
                if "RUN python ${APPDIR}/profile_endpoints.py" not in line
            ]

        with open(os.path.join(self._path, output), "w") as fp:
            fp.writelines(contents)

    def remove_patched_docker_file(self, file_path: str = TESTING_DOCKERFILE):
        os.remove(os.path.join(self._path, file_path))

    def build_image(self, dockerfile: str = TESTING_DOCKERFILE) -> str:
        image = self._client.images.build(
            path=self._path, dockerfile=dockerfile, tag="taktile-cli-test"
        )
        return image[0].id

    def test_import(self, image_id: str):
        container = self._client.containers.run(
            image_id, "python -c 'from src.endpoints import tktl'", detach=True
        )
        status = container.wait()
        return status, container.logs()

    def test_unittest(self, image_id: str):
        container = self._client.containers.run(
            image_id, "python -m pytest ./user_tests/", detach=True
        )
        status = container.wait()
        return status, container.logs()

    def test_profiling(self, image_id: str):
        container = self._client.containers.run(
            image_id, "python /app/profile_endpoints.py", detach=True
        )
        status = container.wait()
        return status, container.logs()

    def test_integration(self, image_id: str):
        container = self._client.containers.run(
            image_id, detach=True, ports={"80/tcp": 8080}
        )
        try:
            for _ in range(5):
                try:
                    time.sleep(5)
                    resp = requests.get("http://localhost:8080/docs")
                    return resp, container.logs()
                except RequestException:
                    pass

            return None, container.logs()
        finally:
            container.kill()



import click
import click_completion
import requests

from tktl.cli.common import ClickGroup
from tktl.cli.delete import delete as delete_commands
from tktl.cli.get import get as get_commands
from tktl.cli.init import init
from tktl.cli.login import login, logout
from tktl.cli.validate import validate
from tktl.commands.version import get_version
from tktl.core.config import settings
from tktl.core.exceptions import ApplicationError, TaktileSdkError
from tktl.core.loggers import CliLogger

click_completion.init()


class TaktileGroup(ClickGroup):
    def main(self, *args, **kwargs):
        try:
            super(TaktileGroup, self).main(*args, **kwargs)
        except requests.exceptions.RequestException as e:
            msg = (
                "Can't connect to Taktile API. "
                "Please check https://status.taktile.com/ for more information."
            )
            CliLogger().error(msg)
        except (ApplicationError, TaktileSdkError) as e:
            if settings.DEBUG:
                raise

            CliLogger().error(e)


@click.group(cls=TaktileGroup, **settings.HELP_COLORS_DICT)
def cli():
    pass


cli.add_command(get_version)
cli.add_command(logout)
cli.add_command(login)
cli.add_command(init)
cli.add_command(get_version)
cli.add_command(delete_commands)
cli.add_command(get_commands)
cli.add_command(validate)
